/**
 ##############################################################################
 # Copyright (C) 2006-2010 INRIA
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
 # License as published by the Free Software Foundation; either
 # version 2 of the License, or (at your option) any later version.
 #
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ##############################################################################
 * 
 */
package fr.inria.icocoa.servicemodel.matching;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;

import fr.inria.icocoa.common.JenaStarter;
import fr.inria.icocoa.common.Tools;
import fr.inria.icocoa.exception.SemanticTypeNotKnowException;
import fr.inria.icocoa.servicemodel.Capability;
import fr.inria.icocoa.servicemodel.IO;
import fr.inria.icocoa.servicemodel.Service;
import fr.inria.icocoa.servicemodel.UserTask;

/**
 * this abstract class gather some method and field that will be shared by all
 * SemanticMatcher.
 * 
 * @author <a href="mailto:sebastien.bianco@inria.fr">sebastien bianco</a> - <a
 *         href="http://www-rocq.inria.fr/arles">Arles Project - INRIA</a>
 * 
 * Creation Date : Sep 11, 2007 9:42:03 AM
 */
public abstract class SemanticMatcher extends Matcher {

	/**
	 * the global ontology use for matching. This ontology has to be the same
	 * for all concept, because we suppose that concept from different ontology
	 * cannot be matched
	 */
	protected OntModel globalOntology = null;
	
	/**
	 * The list of ontologies already read
	 */
	protected List<String> alreadyRed = new ArrayList<String>();

	/**
	 * Load the ontolgy and return a reference on it
	 * 
	 * @param type
	 *            the concept that will give the ontology that will be loaded.
	 *            The ontology is load on the first call of this method, and
	 *            this ontology will be the one that will be used during the all
	 *            matching process. To change it, you must create a new Matcher.
	 * @return the globalOntology or <b>null</b> if the given type is not
	 *         extract from the ontology that has been already loaded.
	 */
	protected OntModel getGlobalOntology(String type) {
		
		String ontologyType = Tools.getFirstPart(type);
		
		if (this.globalOntology == null) {
			// use Jena with a reasoner
			this.globalOntology = JenaStarter.getInstanceOfOntModel(true);
			this.globalOntology.read(JenaStarter.getOntologyDescriptionFile(ontologyType));
			this.alreadyRed.add(ontologyType);
		}else{
			if(!this.alreadyRed.contains(ontologyType)){
				this.globalOntology.read(JenaStarter.getOntologyDescriptionFile(ontologyType));
				this.alreadyRed.add(ontologyType);
			}
		}
		return this.globalOntology;
	}

	/**
	 * Test if two concept are equals or equivalent
	 * 
	 * @param type1
	 *            the first concept
	 * @param type2
	 *            the second concept
	 * @return <i>true</i> if the two concept are equivalent, <i>false</i>
	 *         otherwise.
	 * @throws SemanticTypeNotKnowException
	 *             if one of the two concept is not found in the ontology
	 */
	protected boolean isSemanticlyEquivalent(String type1, String type2)
			throws SemanticTypeNotKnowException {

		/* Test at syntaxic level */
		if (Tools.compareString(type1, type2))
			return true;
		else {

			/* test at semantic level */
			String ontologyType1 = Tools.getFirstPart(type1);
			String ontologyType2 = Tools.getFirstPart(type2);

			// test if the same ontology
			/*if (!ontologyType1.equals(ontologyType2))
				return false;*/

			// get ontology and check if both concept are in correct ontology
			OntModel ontology = this.getGlobalOntology(ontologyType1);
			ontology = this.getGlobalOntology(ontologyType2);
			if (ontology == null)
				throw new SemanticTypeNotKnowException("The type " + type1
						+ " and " + type2
						+ " are not on the ontology used for matching ("
						+ ontology.getProfile().getLabel() + ")");

			// get Ontology class
			OntClass type1Class = ontology.getOntClass(type1);
			OntClass type2Class = ontology.getOntClass(type2);

			// test if the classes exist
			if (type1Class == null)
				throw new SemanticTypeNotKnowException("The type " + type1
						+ " has not been found in the global Ontology");
			if (type2Class == null)
				throw new SemanticTypeNotKnowException("The type " + type2
						+ " has not been found in the global Ontology");

			// ///test the matching between the two types
			boolean result = this.isSemanticlyEquivalent(type1Class, type2Class);

			return result;
		}
	}

	/**
	 * Test if two concept are equals or equivalent
	 * 
	 * @param type1
	 *            the first concept
	 * @param type2
	 *            the second concept
	 * @return <i>true</i> if the two concept are equivalent, <i>false</i>
	 *         otherwise.
	 */
	protected boolean isSemanticlyEquivalent(OntClass type1, OntClass type2) {
		return type2.isSameAs(type1) || type2.hasEquivalentClass(type1) || type1.hasEquivalentClass(type2);
	}

	/**
	 * List the equivalent concept of the one given as parameter. This one is
	 * add at the beginning of the list
	 * 
	 * @param type
	 *            the complete type of the concept that we which an equivalent
	 *            (ontology#concept)
	 * @return the list of equibalent type, starting with the one given as
	 *         parameter
	 * @throws SemanticTypeNotKnowException
	 *             if the type has not been found in the ontology
	 */
	protected List<String> listEquivalentConcept(String type)
			throws SemanticTypeNotKnowException {
		// initialisation
		List<String> result = new ArrayList<String>();
		// result.add(type);

		// get Ontology class
		OntModel ontology = this.getGlobalOntology(type);
		if (ontology == null)
			throw new SemanticTypeNotKnowException("The type " + type
					+ " is not on the ontology used for matching ("
					+ ontology.getProfile().getLabel() + ")");

		OntClass typeClass = ontology.getOntClass(type);

		// test if the classes exist
		if (typeClass == null)
			throw new SemanticTypeNotKnowException("The type " + type
					+ " has not been found in the global Ontology");

		// list all Equivalent type
		Iterator<OntClass> it = this.listEquivalentConcept(typeClass).iterator();

		// walk through list of equivalent concept
		while (it.hasNext()) {
			OntClass next = it.next();
			result.add(next.getLocalName());
			// result.add(next.getURI());
		}

		return result;
	}

	/**
	 * List the equivalent concept of the one given as parameter. This one is
	 * add at the beginning of the list
	 * 
	 * @param type
	 *            the complete type of the concept that we which an equivalent
	 *            (ontology#concept)
	 * @return the list of equibalent type, starting with the one given as
	 *         parameter
	 */
	protected List<OntClass> listEquivalentConcept(OntClass type)
			throws SemanticTypeNotKnowException {
		// initialisation
		List<OntClass> result = new ArrayList<OntClass>();

		// list all Equivalent type
		ExtendedIterator it = type.listEquivalentClasses();

		while (it.hasNext()) {
			OntClass next = (OntClass) it.next();
			result.add(next);
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#capabilityMatch(fr.inria.icocoa.servicemodel.Capability,
	 *      fr.inria.icocoa.servicemodel.Capability)
	 */
	@Override
	public abstract CapabilityMatch capabilityMatch(Capability serviceCapa,
			Capability taskCapa) throws SemanticTypeNotKnowException;

	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#ioMatch(java.lang.String,
	 *      java.lang.String)
	 */
	@Override
	public abstract int ioMatch(String serviceIOSemanticType,
			String taskIOsemanticType) throws SemanticTypeNotKnowException;

	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#serviceMatch(fr.inria.icocoa.servicemodel.Service,
	 *      fr.inria.icocoa.servicemodel.UserTask)
	 */
	@Override
	public abstract ServiceMatch serviceMatch(Service service, UserTask userTask)
			throws SemanticTypeNotKnowException;

	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#ioMatch(fr.inria.icocoa.servicemodel.IO,
	 *      fr.inria.icocoa.servicemodel.IO)
	 */
	@Override
	public abstract IOMatch ioMatch(IO serviceIO, IO taskIO)
			throws SemanticTypeNotKnowException;
}
