/**
 ##############################################################################
 # Copyright (C) 2006-2010 INRIA
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
 # License as published by the Free Software Foundation; either
 # version 2 of the License, or (at your option) any later version.
 #
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ##############################################################################
 * 
 */
package fr.inria.icocoa.servicemodel.matching;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;

import fr.inria.icocoa.common.Tools;
import fr.inria.icocoa.exception.SemanticTypeNotKnowException;
import fr.inria.icocoa.servicemodel.Capability;
import fr.inria.icocoa.servicemodel.IO;
import fr.inria.icocoa.servicemodel.Input;
import fr.inria.icocoa.servicemodel.Output;
import fr.inria.icocoa.servicemodel.Service;
import fr.inria.icocoa.servicemodel.UserTask;
import fr.inria.icocoa.servicemodel.semantictype.SemanticType;

/**
 * 
 * This class implements a Matcher that use synthaxic then semantic to match
 * components (allows less inputs than describe in the user task, and more
 * output). <br>
 * The ontology that will be used is load thanks the first semantic type found
 * during the matching process, and this ontology will be the one that will be
 * used during the all matching process. To change it, you must create a new
 * Matcher. <br>
 * <br>
 * This matcher, that use Jena to access and reason on the ontologies, is abble
 * to produce match with type SUBSUM or PLUGIN (and of course EXACT) <br>
 * 
 * 
 * 
 * @author <a href="mailto:sebastien.bianco@inria.fr">sebastien bianco</a> - <a
 *         href="http://www-rocq.inria.fr/arles">Arles Project - INRIA</a>
 * 
 * Creation Date : Sep 11, 2007 9:35:15 AM
 */
public class WeakSemanticMatcher extends SemanticMatcher {

	/**
	 * The radius of search for concept to determine subsum or plugin matching
	 */
	private int searchRadius = 0;

	/**
	 * Create a new Matcher abble to treat exact, plugin or subsum semantic
	 * match. The default search radius used is 3.
	 * 
	 * 
	 */
	public WeakSemanticMatcher() {
		this(3);
	}

	/**
	 * Create a new Matcher abble to treat exact, plugin or subsum semantic
	 * match
	 * 
	 * @param searchRadius
	 *            the maximum of level in ontologies that will be walk through
	 *            to determine if a concept in plugin or subsum of another one.
	 */
	public WeakSemanticMatcher(int searchRadius) {
		this.searchRadius = searchRadius;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#capabilityMatch(fr.inria.icocoa.servicemodel.Capability,
	 *      fr.inria.icocoa.servicemodel.Capability)
	 */
	@Override
	public CapabilityMatch capabilityMatch(Capability serviceCapa,
			Capability taskCapa) throws SemanticTypeNotKnowException {
		// increase the counter of Capabilities matched
		stats[1]++;

		// get Type
		String serviceCapaSemanticType = serviceCapa.getType();
		String taskCapaSemanticType = taskCapa.getType();

		// test relations
		if (this.isSemanticlyEquivalent(serviceCapaSemanticType,
				taskCapaSemanticType))
			return this.generateMatching(serviceCapa, taskCapa,
					MatchingType.EXACT);
		if (this.isSubsum(serviceCapaSemanticType, taskCapaSemanticType))
			return this.generateMatching(serviceCapa, taskCapa,
					MatchingType.SUBSUMES);
		if (this.isPlugin(serviceCapaSemanticType, taskCapaSemanticType))
			return this.generateMatching(serviceCapa, taskCapa,
					MatchingType.PLUGIN);
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#ioMatch(java.lang.String,
	 *      java.lang.String)
	 */
	@Override
	public int ioMatch(String serviceIOSemanticType, String taskIOsemanticType)
			throws SemanticTypeNotKnowException {
		// increase the counter of IO matched
		stats[2]++;

		return this.typeMatch(serviceIOSemanticType, taskIOsemanticType);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#typeMatch(fr.inria.icocoa.servicemodel.semantictype.SemanticType,
	 *      fr.inria.icocoa.servicemodel.semantictype.SemanticType)
	 */
	@Override
	public short typeMatch(SemanticType serviceType, SemanticType taskType) throws SemanticTypeNotKnowException {
		return this
				.typeMatch(serviceType.getType(), taskType.getType());
	}

	/**
	 * Get a matching between the two concepts (the task Concept comparing to
	 * the service one)
	 * 
	 * @param serviceTypeName
	 *            the name of the reference concept
	 * @param taskTypeName
	 *            the name of the concept to compare with
	 * @return
	 *            <ul>
	 *            <li>MatchingType.EXACT if the two concept are equals or
	 *            equivalent</li>
	 *            <li>Matching.SUBSUMES if task concept subsums the service one
	 *            </li>
	 *            <li>Matching.PLUGIN if service concept subsums the task one
	 *            </li>
	 *            <li>Matching.FAIL if there is no link between the two concept</li>
	 *            </ul>
	 * @throws SemanticTypeNotKnowException
	 */
	private short typeMatch(String serviceTypeName, String taskTypeName)
			throws SemanticTypeNotKnowException {
		if (this.isSemanticlyEquivalent(serviceTypeName, taskTypeName))
			return MatchingType.EXACT;
		if (this.isSubsum(serviceTypeName, taskTypeName))
			return MatchingType.SUBSUMES;
		if (this.isPlugin(serviceTypeName, taskTypeName))
			return MatchingType.PLUGIN;
		return MatchingType.FAIL;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#serviceMatch(fr.inria.icocoa.servicemodel.Service,
	 *      fr.inria.icocoa.servicemodel.UserTask)
	 */
	@Override
	public ServiceMatch serviceMatch(Service service, UserTask userTask)
			throws SemanticTypeNotKnowException {
		// increase the counter of Services matched
		stats[0]++;

		ServiceMatch result = this.partialExactMatch(service, userTask);

		// if no matching
		if (result == null)
			return null;

		// test the number of Functionality matching (all required need to me
		// matched)
		if (result.getCapabilityMatches().size() < (userTask
				.getRequired_Capabilities().size() + userTask
				.getCapabilitiesMap().size()))
			return null;

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.matching.Matcher#ioMatch(fr.inria.icocoa.servicemodel.IO,
	 *      fr.inria.icocoa.servicemodel.IO)
	 */
	@Override
	public IOMatch ioMatch(IO serviceIO, IO taskIO)
			throws SemanticTypeNotKnowException {
		if (serviceIO.getClass().equals(taskIO.getClass())) {
			switch (this.ioMatch(serviceIO.getType(), taskIO.getType())) {
			case MatchingType.FAIL:
				return null;
			case MatchingType.EXACT:
				return new IOMatch(taskIO, serviceIO, MatchingType.EXACT);
			case MatchingType.SUBSUMES:
				return new IOMatch(taskIO, serviceIO, MatchingType.SUBSUMES);
			case MatchingType.PLUGIN:
				return new IOMatch(taskIO, serviceIO, MatchingType.PLUGIN);
			default:
				return null;
			}
		}
		return null;
	}

	/**
	 * Test if a type is plugin of another type
	 * 
	 * @param type1
	 *            the reference type
	 * @param type2
	 *            the type to test
	 * @return <i>true</i> if type2 is plugin of type1; <i>false</i>
	 *         otherwise.
	 * @throws SemanticTypeNotKnowException
	 *             if a type is not found in the Ontology
	 */
	private boolean isPlugin(String type1, String type2)
			throws SemanticTypeNotKnowException {
		return this.isSubsum(type2, type1);
	}

	/**
	 * Test if a type is subsum of another type
	 * 
	 * @param type1
	 *            the reference type
	 * @param type2
	 *            the type to test
	 * @return <i>true</i> if type2 is subsum of type1; <i>false</i>
	 *         otherwise.
	 * @throws SemanticTypeNotKnowException
	 *             if a type is not found in the Ontology
	 */
	private boolean isSubsum(String type1, String type2)
			throws SemanticTypeNotKnowException {

		/*
		 * Initialise Ontology object
		 */

		// get ontology URL of concepts
		String ontologyType1 = Tools.getFirstPart(type1);
		String ontologyType2 = Tools.getFirstPart(type2);

		// test if the same ontology
		if (!ontologyType1.equals(ontologyType2))
			return false;

		// get ontology and check if both concept are in correct ontology
		OntModel ontology = this.getGlobalOntology(ontologyType1);
		if (ontology == null)
			throw new SemanticTypeNotKnowException("The type " + type1
					+ " and " + type2
					+ " are not on the ontology used for matching ("
					+ ontology.getProfile().getLabel() + ")");

		// get Ontology class
		OntClass type1Class = ontology.getOntClass(type1);
		OntClass type2Class = ontology.getOntClass(type2);

		// test if the classes exist
		if (type1Class == null)
			throw new SemanticTypeNotKnowException("The type " + type1
					+ " has not been found in the global Ontology");
		if (type2Class == null)
			throw new SemanticTypeNotKnowException("The type " + type2
					+ " has not been found in the global Ontology");

		/*
		 * Search for Concept
		 */

		// initialize list (FIFO) of class to explore
		List<QueueElement> toTest = new ArrayList<QueueElement>();
		toTest.add(new QueueElement(type1Class, 0));

		// search of the concept by increasing range of search
		QueueElement current = null;
		while (!toTest.isEmpty()) {

			// DebugPrinter.print(toTest,"toTest",DebugPrinter.INFO);

			// get the head of the queue
			current = toTest.remove(0);

			// DebugPrinter.print("current = "+current,DebugPrinter.INFO);

			// ignore first level otherwise equivalent to method
			// isSemanticlyEquivalent
			if (current.getLevel() > 0) {
				// get list of equivalent concept
				List<OntClass> equivalents = this.listEquivalentConcept(current
						.getElement());

				// walk through the list of equivalents
				Iterator<OntClass> it = equivalents.iterator();
				OntClass equivalent = null;
				while (it.hasNext()) {
					equivalent = it.next();
					if (this.isSemanticlyEquivalent(equivalent, type2Class))
						return true;
				}
			}

			/*
			 * if execution is here, it means that the type1Concept has not been
			 * found with this element, so add subclasses of this concept
			 */

			// test if the maximum radius has been reached
			if (current.getLevel() < this.getSearchRadius()) {

				// add all children in the list of class to explore
				ExtendedIterator childrenIterator = current.getElement()
						.listSubClasses(true);
				QueueElement toAdd = null;
				while (childrenIterator.hasNext()) {
					toAdd = new QueueElement(
							(OntClass) childrenIterator.next(), current
									.getLevel() + 1);
					// DebugPrinter.print(">>> toAdd =
					// "+toAdd,DebugPrinter.INFO);
					toTest.add(toAdd);

				}
			}
		}

		/*
		 * If exectution is here, it means that the concept type2 has not be
		 * found in subconcept of type1 => type2 is not a subsum of type1
		 */
		return false;
	}

	/**
	 * Generate the complete FunctionalityMatching object corresponding to the
	 * matching between this Capability and the one give as parameter
	 * 
	 * @param serviceCapa
	 *            the current Capability
	 * @param taskCapa
	 *            the capability to match with.
	 * @param matchingType
	 *            the type of matching (cf MatchingType)
	 * @return a MatchingCapabibility Object containing all information, if the
	 *         the matching is possible, <i>null</i> otherwise.
	 * @throws SemanticTypeNotKnowException
	 *             if one type is not found
	 */
	private CapabilityMatch generateMatching(Capability serviceCapa,
			Capability taskCapa, short matchingType)
			throws SemanticTypeNotKnowException {

		// generate IOMatching
		List<IOMatch> iomatching = this.ioMatch(serviceCapa, taskCapa);

		// if no io matching
		if (iomatching == null)
			return null;

		// create the matching at Capability level
		CapabilityMatch result = new CapabilityMatch(taskCapa, serviceCapa,
				matchingType);

		// add matching at io level
		for (Iterator<IOMatch> it = iomatching.iterator(); it.hasNext();)
			result.addIOMatching(it.next());

		return result;
	}

	/**
	 * Try to match the IO of the given capabilities with this (used as UserTask
	 * capability)
	 * 
	 * @param serviceCapa :
	 *            the current Capability
	 * @param taskCapa :
	 *            the capability to match with.
	 * @return the List of matching (at IO level), <i>null</i> if we can't
	 *         match (not the the same type or not the same number)
	 * @throws SemanticTypeNotKnowException
	 *             if one type is not found
	 */
	private List<IOMatch> ioMatch(Capability serviceCapa, Capability taskCapa)
			throws SemanticTypeNotKnowException {
		// Map<IO, IO> matchs = new HashMap<IO, IO>();

		List<IOMatch> listOfMatchs = new ArrayList<IOMatch>();

		/*
		 * Match Inputs (the capa (service capa) could have less inputs the
		 * Usertask)
		 */
		if (taskCapa.getInputs().size() <= serviceCapa.getInputs().size()) {

			/* For each input of selected service capability... */
			for (Iterator<Input> it = taskCapa.getInputs().values().iterator(); it
					.hasNext();) {
				Input toFind = it.next();

				// String type = toFind.getType();
				boolean found = false;

				/*
				 * ...Look if there is an corresponding input for task
				 * capability
				 */
				loopServiceInput: for (Iterator<Input> it2 = serviceCapa
						.getInputs().values().iterator(); !found
						&& it2.hasNext();) {
					Input current = it2.next();

					// test matching between task IO "toFind" and service IO
					// "current"
					IOMatch tmpMatch = this.ioMatch(current, toFind);
					if (tmpMatch != null) {
						listOfMatchs.add(tmpMatch);

						// stop the loop to speed up the process
						found = true;
						break loopServiceInput;
					}

				}
				if (!found) {
					return null;
				}
			}
		} else {
			return null;
		}

		/*
		 * Match output (the capa (service capa) could have more outputs the
		 * Usertask)
		 */
		if (taskCapa.getOutputs().size() >= serviceCapa.getOutputs().size()) {

			/* For each output of usertask capability... */
			for (Iterator<Output> it = serviceCapa.getOutputs().values()
					.iterator(); it.hasNext();) {
				Output toFind = it.next();

				// String type = toFind.getType();
				boolean found = false;
				/*
				 * ...Look if there is an corresponding input for selected
				 * service capability
				 */
				loopTaskOutput: for (Iterator<Output> it2 = taskCapa
						.getOutputs().values().iterator(); it2.hasNext()
						&& !found;) {
					Output current = it2.next();

					// test matching between service IO "toFind" and task IO
					// "current"
					IOMatch tmpMatch = this.ioMatch(toFind, current);
					if (tmpMatch != null) {
						listOfMatchs.add(tmpMatch);

						// stop the loop to speed up the process
						found = true;
						break loopTaskOutput;
					}

				}
				if (!found) {
					return null;
				}
			}
		} else {
			return null;
		}
		// return matchs;
		return listOfMatchs;
	}

	/**
	 * Get a partial matching between the task and the current service
	 * 
	 * @param service :
	 *            the current service
	 * @param userTask :
	 *            the task to match with
	 * @return a ServiceMatch object containing the matching information; if
	 *         some capabilities could be matched, <i>null</i> if no matching
	 *         is possible
	 * @throws SemanticTypeNotKnowException
	 *             if a type is not knowed
	 */
	private ServiceMatch partialExactMatch(Service service, UserTask userTask)
			throws SemanticTypeNotKnowException {

		ServiceMatch result = new ServiceMatch(userTask, service,
				MatchingType.EXACT);

		/* Get the required capabilities */
		List<Capability> required = userTask.getRequiredCapabilities();

		// check all required capability
		for (Iterator<Capability> it = required.iterator(); it.hasNext();) {
			Capability required_capa = it.next();

			List<String> equivalents = this.listEquivalentConcept(required_capa
					.getType());

			for (String equivalent : equivalents) {
				// match with all capability with the same/equivalent type
				List<Capability> sameType = service
						.getCapabilityFromTypeName(equivalent);

				// try to match at CapabilityLevel (<=> test io)
				for (Iterator<Capability> it2 = sameType.iterator(); it2
						.hasNext();) {
					Capability service_capa = it2.next();
					CapabilityMatch cm = service_capa
							.match(required_capa, this);
					if (cm != null)
						result.addCapabilityMatch(cm);
				}
			}
		}

		/* Get the other capabilities */
		List<Capability> subcapabilties = userTask.getCapabilities();

		// check other capability
		for (Iterator<Capability> it = subcapabilties.iterator(); it.hasNext();) {
			Capability subcapa = it.next();

			List<String> equivalents = this.listEquivalentConcept(subcapa
					.getType());

			for (String equivalent : equivalents) {
				// match with all capability with the same/equivalent type
				List<Capability> sameType = service
						.getCapabilityFromTypeName(equivalent);

				// try to match at CapabilityLevel (<=> test io)
				for (Iterator<Capability> it2 = sameType.iterator(); it2
						.hasNext();) {
					Capability service_capa = it2.next();
					CapabilityMatch cm = service_capa.match(subcapa, this);
					if (cm != null)
						result.addCapabilityMatch(cm);
				}
			}
		}

		// if no matching
		if (result.getCapabilityMatches().size() == 0)
			return null;

		return result;
	}

	/**
	 * @return the searchRadius
	 */
	public int getSearchRadius() {
		return searchRadius;
	}

	/**
	 * @param searchRadius
	 *            the searchRadius to set
	 */
	public void setSearchRadius(int searchRadius) {
		this.searchRadius = searchRadius;
	}

	/**
	 * This class permit to store both a OntClass object and the gap between
	 * this class and the one we are currently testing
	 * 
	 * @author <a href="mailto:sebastien.bianco@inria.fr">sebastien bianco</a> -
	 *         <a href="http://www-rocq.inria.fr/arles">Arles Project - INRIA</a>
	 * 
	 * Creation Date : Sep 11, 2007 2:51:11 PM
	 */
	private class QueueElement {

		/** the object that we want to sore in the Queue */
		private OntClass element = null;

		/** The current length of the element in the ontology */
		private int level = 0;

		/**
		 * @param element
		 * @param level
		 */
		public QueueElement(OntClass element, int level) {
			this.element = element;
			this.level = level;
		}

		/**
		 * @return the element
		 */
		public OntClass getElement() {
			return element;
		}

		/**
		 * @param element
		 *            the element to set
		 */
		public void setElement(OntClass element) {
			this.element = element;
		}

		/**
		 * @return the level
		 */
		public int getLevel() {
			return level;
		}

		/**
		 * @param level
		 *            the level to set
		 */
		public void setLevel(int level) {
			this.level = level;
		}

		@Override
		public String toString() {
			return this.getLevel() + " <---> "
					+ this.getElement().getLocalName();
		}

	}

}
