/**
 ##############################################################################
 # Copyright (C) 2006-2010 INRIA
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
 # License as published by the Free Software Foundation; either
 # version 2 of the License, or (at your option) any later version.
 #
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ##############################################################################
 * 
 */
package fr.inria.icocoa.servicemodel.parsers.jenaparser;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

import fr.inria.icocoa.common.DebugPrinter;
import fr.inria.icocoa.common.PropertyManager;

/**
 * WSDLParser.java<br>
 * This class will parse a WSDL file to extract some information
 * 
 * @author <a href="mailto:sebastien.bianco@inria.fr">sebastien bianco</a> - <a
 *         href="http://www-rocq.inria.fr/arles">Arles Project - INRIA</a>
 * 
 * Creation Date : Feb 6, 2007 10:05:34 AM
 */
public class WSDLParser {

	/** The service uri read from the WSDL definition */
	private String myuri = "";

	/** the service name read from the WSDL file */
	private String myServiceName = "";

	/** the list of messages contained into the WSDL file */
	private List<String> myMessage = new ArrayList<String>();

	/** the service port type read from the WSDL */
	private String myPortType = "";

	/** the service port read from the WSDL */
	private String myPort = "";

	/** the real url of the service */
	private String soapAddress = "";

	/**
	 * The property manager used by this class
	 */
	private PropertyManager propertyManager = PropertyManager.getInstance();

	/**
	 * Constructor of the class. Parse the process and fill the different field
	 * 
	 * @param wsdl_content
	 *            WSDL Description of the service
	 * @throws IOException
	 *             if a problem occurs with the given file
	 * @throws SAXException
	 *             if a problem occurs during the parsing of the given file
	 */
	public WSDLParser(String wsdl_content) throws IOException, SAXException {
		this.parseWSDL(wsdl_content);
	}

	/**
	 * Set the URI and the service name of the service (extact information from
	 * WSDL document)
	 * 
	 * @param wsdl_content
	 *            WSDL Description of the service
	 * @throws IOException
	 *             if a problem occurs with the given file
	 * @throws SAXException
	 *             if a problem occurs during the parsing of the given file
	 */
	private void parseWSDL(final String wsdl_content) throws IOException,
			SAXException {

		XMLReader myReader = XMLReaderFactory.createXMLReader();

		// creation of the SAX handler
		DefaultHandler handler = new DefaultHandler() {

			/** Indicate if we enter the balise "service" */
			private boolean serviceTagFound = false;

			@Override
			public void startElement(String uri, String localName,
					String qName, Attributes attributes) throws SAXException {

				// get the targetNamespace attribute of the definitions tag
				if (localName.equals(propertyManager.WSDL_DEFINITIONS_TAG)) {

					try {
						myuri = attributes.getValue(propertyManager.WSDL_TARGETNAMESPACE_PROPERTY);

						// myuri must not finsih with a "/", in caseof we
						// suppress it
						/*
						 * if (myuri.endsWith("/")) { myuri = myuri.substring(0,
						 * myuri.length() - 1); }
						 */
					} catch (Exception e) {
						throw new SAXException("no target Namespace in WSDL");
					}

				} else if (localName.equals(propertyManager.WSDL_SERVICE_TAG)) {

					try {
						myServiceName = attributes.getValue(propertyManager.WSDL_NAME_PROPERTY);
						serviceTagFound = true;
					} catch (Exception e) {
						throw new SAXException("no service name in WSDL");
					}

				} else if (localName.equals(propertyManager.WSDL_MESSAGE_TAG)) {

					try {
						myMessage.add(attributes.getValue(propertyManager.WSDL_NAME_PROPERTY));
					} catch (Exception e) {
						throw new SAXException("no service name in WSDL");
					}

				} else if (localName.equals(propertyManager.WSDL_PORT_TAG)
						&& serviceTagFound) {

					try {
						myPort = attributes.getValue(propertyManager.WSDL_NAME_PROPERTY);
					} catch (Exception e) {
						throw new SAXException("no port type in WSDL");
					}

				} else if (localName.equals(propertyManager.WSDL_PORTTYPE_TAG)) {

					try {
						myPortType = attributes.getValue(propertyManager.WSDL_NAME_PROPERTY);
					} catch (Exception e) {
						throw new SAXException("no port type in WSDL");
					}

				} else if (localName.equals(propertyManager.WSDL_SOAPADDRESS_TAG)) {
					try {
						soapAddress = attributes.getValue(propertyManager.WSDL_LOCATION_PROPERTY);
					} catch (Exception e) {
						throw new SAXException("no port type in WSDL");
					}
				}

			}

			/*
			 * (non-Javadoc)
			 * 
			 * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String,
			 *      java.lang.String, java.lang.String)
			 */
			@Override
			public void endElement(String uri, String localName, String qName)
					throws SAXException {
				super.endElement(uri, localName, qName);

				if (localName.equals(propertyManager.WSDL_SERVICE_TAG)) {
					serviceTagFound = false;
				}
			}

		};

		// put the handler in the sax machine
		myReader.setContentHandler(handler);
		myReader.setErrorHandler(handler);

		// parse the String
		InputStream is = new java.io.ByteArrayInputStream(
				wsdl_content.getBytes());
		myReader.parse(new InputSource(is));

	}

	/**
	 * @return the myMessage
	 */
	public List<String> getMyMessage() {
		return myMessage;
	}

	/**
	 * @return the myPort
	 */
	public String getMyPort() {
		return myPort;
	}

	/**
	 * @return the myPortType
	 */
	public String getMyPortType() {
		return myPortType;
	}

	/**
	 * @return the myServiceName
	 */
	public String getMyServiceName() {
		return myServiceName;
	}

	/**
	 * @return the myuri
	 */
	public String getMyuri() {
		return myuri;
	}

	/**
	 * @return the propertyManager
	 */
	public PropertyManager getPropertyManager() {
		return propertyManager;
	}

	/**
	 * @return the soapAddress
	 */
	public String getSoapAddress() {
		return soapAddress;
	}

	@Override
	public String toString() {
		Class c = this.getClass();
		Field[] f = c.getDeclaredFields();
		String result = "WSDLParser = " + f.length;
		for (int i = 0; i < f.length; i++) {
			try {
				result += "\n\t> " + f[i].getName() + " = " + f[i].get(this);
			} catch (IllegalArgumentException e) {
				DebugPrinter.print(e.getStackTrace(), e.getMessage(), 2);
			} catch (IllegalAccessException e) {
				DebugPrinter.print(e.getStackTrace(), e.getMessage(), 2);
			}
		}
		return result;
	}

	public List<String> getMessages() {
		return this.myMessage;
	}

}
