/**
 ##############################################################################
 # Copyright (C) 2006-2010 INRIA
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
 # License as published by the Free Software Foundation; either
 # version 2 of the License, or (at your option) any later version.
 #
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ##############################################################################
 * 
 */
package fr.inria.icocoa.servicemodel.parsers.jenaparser.tools;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import zeus.util.Tree;
import zeus.util.TreeNode;
import fr.inria.icocoa.servicemodel.Capability;

/**
 * CapabilityForest.java<br>
 * This class permit to manage the trees of process during the parsing. This
 * tree will be usefull for maintains the databinding.
 * 
 * @author <a href="mailto:sebastien.bianco@inria.fr">sebastien bianco</a> - <a
 *         href="http://www-rocq.inria.fr/arles">Arles Project - INRIA</a>
 * 
 * Creation Date : Jul 18, 2007 4:33:06 PM
 */
public class CapabilityForest {

	/** the list of tree (-> one tree = one composite capability) */
	private List<Tree<Capability>> forest = new ArrayList<Tree<Capability>>();

	/** the map CapablityLocalName -> TreeNode */
	private Map<String, TreeNode<Capability>> capabilitiesNode = new HashMap<String, TreeNode<Capability>>();

	/**
	 * Add Capability as Basic Capability (don't use the hierchy of composite
	 * Capability to design automaticly the tree)
	 * 
	 * @param parent
	 *            the parent Capability of the "child" => parent is a composite
	 *            capability. This could be <b>null</b>, in that case, a new
	 *            tree will be add with "child" as root.
	 * @param child
	 *            the Capability we want to add
	 * @return the TreeNode associated with "child" object, <b>null</b> if the
	 *         operation doesn't succeed
	 */
	public TreeNode<Capability> addSubCapability(Capability parent,
			Capability child) {
		try {
			// if the Capability has no parent
			if (parent != null) {

				// find the Node assosciated to the parent
				TreeNode<Capability> finded = this.capabilitiesNode.get(parent.getLocalName());

				/*
				 * if the parent node has been find add a new node with child
				 * Capability in the list of child node of the finded one
				 */
				if (finded != null) {
					/*
					 * search the node in the forest, if the node is contained,
					 * just add child, else add it as a root of a new tree
					 */
					boolean getIt = false;
					for (int i = 0; i < this.forest.size() && !getIt; i++) {
						Tree<Capability> tmp = forest.get(i);
						getIt |= tmp.containsValue(parent);
					}
					if (!getIt) {
						// add the new child to the parent node
						TreeNode<Capability> added = finded.addChild(child);

						// add the new TreeNode in the Map
						this.capabilitiesNode.put(child.getLocalName(), added);

						// add the parent node as a root of a tree
						this.forest.add(new Tree<Capability>(finded));

						return added;
					} else {
						// add the new child to the parent node
						TreeNode<Capability> added = finded.addChild(child);

						// add the new TreeNode in the Map
						this.capabilitiesNode.put(child.getLocalName(), added);

						return added;
					}

				} else {
					/*
					 * if the parent node has not been found, juste create the
					 * node assocate to the parent, add a child to this node
					 * associate to the "child", and add both into the map
					 */
					TreeNode<Capability> nodeP = new TreeNode<Capability>(
							parent);

					// add the new child to the parent node
					TreeNode<Capability> nodeC = nodeP.addChild(child);

					// add the two new TreeNode in the Map
					this.capabilitiesNode.put(parent.getLocalName(), nodeP);
					this.capabilitiesNode.put(child.getLocalName(), nodeC);

					return nodeC;
				}

			} else {

				if (child != null) {
					// find the Node assosciated to the child
					TreeNode<Capability> finded = this.capabilitiesNode.get(child.getLocalName());

					/*
					 * if the node has no parent, then create a new tree which
					 * root is parent Capability, and add the child node as the
					 * first child of the parent node
					 */
					if (finded == null) {

						Tree<Capability> tree = new Tree<Capability>(child);
						forest.add(tree);

						TreeNode<Capability> root = tree.getRoot();

						// add the new TreeNode in the Map
						this.capabilitiesNode.put(child.getLocalName(), root);

						// add the new TreeNode in the Map
						return root;
					} else {
						/*
						 * if the node already exist, then just add is as a root
						 * of a new tree
						 */
						Tree<Capability> tree = new Tree<Capability>(finded);
						forest.add(tree);

						return finded;
					}
				} else
					return null; // child = parent = null => nothing to do !
			}
		} catch (IllegalArgumentException e) {
			return null;
		}
	}

	/**
	 * Add a new Capability to the forest. The given Capability is put as a new
	 * tree of the Forest. If the Capability is composite, all the
	 * subcapabilities are added to this tree
	 * 
	 * @param c
	 */
	public void addCapability(Capability c) {
		if (c != null) {
			// add c as a new root subcapability -> create a new tree from c
			this.addSubCapability(null, c);

			// if c is a composite c then add all subcapabilties => walk through
			// the tree
			if (!c.isAtomic()) {
				this.visitChildCapability(c);
			}
		}
	}

	/**
	 * Delete all data from the CapabilityForest
	 */
	public void clear() {
		this.forest.clear();
		this.capabilitiesNode.clear();
	}

	/**
	 * @return the capabilitiesNode
	 */
	public Map<String, TreeNode<Capability>> getCapabilitiesNode() {
		return capabilitiesNode;
	}

	/**
	 * @return the forest
	 */
	public List<Tree<Capability>> getForest() {
		return forest;
	}

	/**
	 * Display the local name of the Capability that compose the forest
	 * 
	 * @return a String representation of the forest, empty String if the forest
	 *         contains no tree.
	 */
	public String displayForest() {
		String result = "";
		for (Tree<Capability> tree : this.forest) {
			result += "++ " + this.diplayTreeNode("   ", tree.getRoot());
		}
		return result;
	}

	/**
	 * Display the local name of the capabilty that compose the branch whose
	 * "node" is the root
	 * 
	 * @param prefix
	 *            the prefix to add before each line to get a good indentation
	 * @param node
	 *            the root of the branch to display
	 * @return a String representation of the branch
	 */
	private String diplayTreeNode(String prefix, TreeNode<Capability> node) {
		String result = "";
		if (node.isTerminal()) {
			result += prefix + "- " + node.getValue().getLocalName() + "\n";
		} else {
			result += prefix + "+ " + node.getValue().getLocalName() + "\n";
			for (TreeNode<Capability> child : node.getChildren()) {
				result += prefix + " "
						+ this.diplayTreeNode(prefix + " ", child);
			}
		}
		return result;
	}

	@Override
	public String toString() {
		return this.displayForest();
	}

	/**
	 * Visit all the subcapabilities of a given Capability, and add it in the
	 * forest
	 * 
	 * @param father
	 *            the capability that we need to visit the children
	 */
	private void visitChildCapability(Capability father) {
		if (father.isAtomic()) {
			// do nothing,
		} else {

			// get node associated to the father
			TreeNode<Capability> fatherNode = this.capabilitiesNode.get(father.getLocalName());

			// if the node has children, it means that the list of child is
			// already know, so we don't need to complete it. If not, we need to
			// walk trough the child and complete the stuff
			if (fatherNode.isTerminal()) {

				// get the list of subcapabilities
				List<Capability> children = father.getSubCapabilities();

				// walk through the list
				for (Capability child : children) {
					if (child.isAtomic()) {
						// add capability in the tree
						this.addSubCapability(father, child);
					} else {

						// add capability in the tree
						this.addSubCapability(father, child);

						// launch recursive visit of all child of this child
						this.visitChildCapability(child);
					}
				}
			}
		}
	}

}
