/**
 ##############################################################################
 # Copyright (C) 2006-2010 INRIA
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
 # License as published by the Free Software Foundation; either
 # version 2 of the License, or (at your option) any later version.
 #
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ##############################################################################
 * 
 */
package fr.inria.icocoa.servicemodel.semantictype.encoder.jenaencoder;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.mindswap.pellet.jena.PelletInfGraph;

import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.vocabulary.OWL;

import fr.inria.icocoa.common.DebugPrinter;
import fr.inria.icocoa.common.JenaStarter;
import fr.inria.icocoa.servicemodel.semantictype.encryptionmanagement.OntologiesGraphGenerator;
import fr.inria.icocoa.servicemodel.semantictype.encryptionmanagement.toolobjects.Node;

/**
 * JenaOntologiesGraphGenerator.java
 * 
 * @author <a href="mailto:sebastien.bianco@inria.fr">sebastien bianco</a> - <a
 *         href="http://www-rocq.inria.fr/arles">Arles Project - INRIA</a>
 * 
 * Creation Date : Mar 31, 2008 3:07:46 PM
 */
public class JenaOntologiesGraphGenerator implements OntologiesGraphGenerator {

	static{
		//DebugPrinter.print("JenaOntologiesGraphGenerator found !", DebugPrinter.INFO);
	}
	
	/**
	 * the OntModel that will be used during the complete process of graph
	 * creation
	 */
	private OntModel ontmodel = null;

	/**
	 * relation between the node and the relative OntClass object. This permit
	 * to avoid that the algorithm never ended.
	 */
	private Map<String, Node> alreadyFoundConcepts = null;
	
	/**
	 * Indicate if the required only direct relation for Graph generation or if we want all inferred relation
	 */
	private boolean getOnlyDirectRelation = false;

	/**
	 * Default constructor
	 *
	 */
	public JenaOntologiesGraphGenerator(){		
		this(false);
		//DebugPrinter.print("new instance of JenaOntologiesGraphGenerator",DebugPrinter.INFO);
	}

	/**
	 * Constructor by giving as parameter the fact that we required or not all inferred relationship
	 * @param getOnlyDirectRelation  Indicate if the required only direct relation for Graph generation or if we want all inferred relation
	 */
	public JenaOntologiesGraphGenerator(boolean getOnlyDirectRelation){
		//DebugPrinter.print("new instance of JenaOntologiesGraphGenerator => getOnlyDirectRelation = "+getOnlyDirectRelation,DebugPrinter.INFO);
		this.getOnlyDirectRelation = getOnlyDirectRelation;
		
		//initialise OntModel
		//DebugPrinter.print("new instance of JenaOntologiesGraphGenerator => initialize ontModel",DebugPrinter.INFO);
		ontmodel = JenaStarter.getInstanceOfOntModel(true);
		
		// initialise the map of concept already found
		//DebugPrinter.print("new instance of JenaOntologiesGraphGenerator => initialize Map",DebugPrinter.INFO);
		alreadyFoundConcepts = new HashMap<String, Node>();

	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see fr.inria.icocoa.servicemodel.semantictype.encoder.jenaencoder.OntologiesGraphGenerator#generateOntologiesGraph(java.util.List)
	 */
	public Node generateOntologiesGraph(List<String> listOfOntologies) {

		// re-initialise the map of concept already found
		this.alreadyFoundConcepts.clear();

		// read all required ontologies
		Iterator<String> it = listOfOntologies.iterator();
		while (it.hasNext())
			ontmodel.read(it.next());

		// create a reasoner
		ontmodel.prepare();
		((PelletInfGraph) ontmodel.getGraph()).getKB().classify();

		// create all node object, from leave to root
		this.generationOfNodes();

		// really create the Graph of Node
		this.createGraph();		
		
		// Get the associated root Node
		return alreadyFoundConcepts.get(OWL.Thing.getURI());

	}

	/**
	 * Create the graph or relation between node
	 * (subclasses/parents/equivalents)
	 * 
	 */
	@SuppressWarnings("unchecked")
	private void createGraph() {
		//DebugPrinter.print(alreadyFoundConcepts, "start graph generation",DebugPrinter.INFO);

		/*
		 * For each Class, create the relations equivalent / subclasses
		 */
		Iterator<String> classIt = alreadyFoundConcepts.keySet().iterator();
		while (classIt.hasNext()) {
			String currentClassURI = classIt.next();
			//DebugPrinter.print("currentClassURI ="+currentClassURI,DebugPrinter.INFO);
			OntClass currentClass = ontmodel.getOntClass(currentClassURI);
			Node currentNode = alreadyFoundConcepts.get(currentClassURI);

			// check that the Node exist
			if (currentNode != null && currentClass!=null) {

				/*
				 * list equivalents
				 */
				Iterator<OntClass> equivalentIt = currentClass
						.listEquivalentClasses();
				while (equivalentIt.hasNext()) {
					OntClass equivalentClass = equivalentIt.next();
					//DebugPrinter.print("an equivalent -> "+equivalentClass.getURI(), DebugPrinter.INFO);
					if (equivalentClass != null
							&& equivalentClass.getURI() != null) {
						if (!equivalentClass.getURI().equalsIgnoreCase(
								OWL.Nothing.getURI())) {
							Node equivalentNode = alreadyFoundConcepts
									.get(equivalentClass.getURI());
							// check that the node exist
							if (equivalentNode != null) {
								equivalentNode.addEquivalent(currentNode);
								currentNode.addEquivalent(equivalentNode);
							} else {
								DebugPrinter
										.print(
												"The node corresponding to the concept "
														+ equivalentClass
																.getURI()
														+ " has not been found ! (equivalent of "
														+ currentClassURI + ")",
												DebugPrinter.INFO);
							}
						}
					}
				}

				/*
				 * list children
				 */
				Iterator<OntClass> childrenIt = currentClass
						.listSubClasses(this.getOnlyDirectRelation);
				while (childrenIt.hasNext()) {
					OntClass childClass = childrenIt.next();
					//DebugPrinter.print("a child -> "+childClass.getURI(), DebugPrinter.INFO);

					if (childClass != null && childClass.getURI() != null) {
						if (!childClass.getURI().equalsIgnoreCase(
								OWL.Nothing.getURI())) {
							Node childNode = alreadyFoundConcepts
									.get(childClass.getURI());
							// check that the node exist
							if (childNode != null) {
								currentNode.addChild(childNode);
							} else {
								DebugPrinter
										.print(
												"The node corresponding to the concept "
														+ childClass.getURI()
														+ " has not been found ! (child of "
														+ currentClassURI + ")",
												DebugPrinter.INFO);
							}
						}/*else{
							DebugPrinter
							.print("Nothing found",	DebugPrinter.INFO);
						}*/
					}
				}

			} else {
				DebugPrinter.print("The node corresponding to the concept "
						+ currentClassURI + " has not been found !",
						DebugPrinter.INFO);
			}
		}
	}

	/**
	 * Create all Node corresponding to graph
	 * 
	 */
	@SuppressWarnings("unchecked")
	private void generationOfNodes() {
		// try {
		// create list of classes without sub-classes
		LinkedList<OntClass> classesToCheck = new LinkedList<OntClass>();

		// initialise list
		Iterator<OntClass> classIt = ontmodel.listClasses();
		while (classIt.hasNext()) {
			classesToCheck.add(classIt.next());
		}

		/*
		 * Check that theses classes have no subclass. If they have unregistered
		 * classes, create associated node and registered them, In any case, add
		 * the subclasses in the list classes to check
		 */
		while (!classesToCheck.isEmpty()) {
			
			
			OntClass currentClass = classesToCheck.remove();
			/*
			 * DebugPrinter.print("classesToCheck - size = " +
			 * classesToCheck.size(), DebugPrinter.INFO);
			 */

			/*
			 * Create current Node
			 */
			if (currentClass != null && currentClass.getURI() != null) {
				Node currentNode = new Node(currentClass.getURI());
				if (!currentClass.getURI().equalsIgnoreCase(
						OWL.Nothing.getURI())) {
					if (!alreadyFoundConcepts
							.containsKey(currentClass.getURI())) {
						alreadyFoundConcepts.put(currentClass.getURI(),
								currentNode);
						//DebugPrinter.print("put "+currentClass.getURI()+" in alreadyFoundConcept" , DebugPrinter.INFO);
					}
				}

				/*
				 * Create subclass node
				 */
				Iterator<OntClass> subclassIt = currentClass
						.listSubClasses(this.getOnlyDirectRelation);
				while (subclassIt.hasNext()) {
					OntClass subClass = subclassIt.next();

					if (subClass != null && subClass.getURI() != null) {
						if (!subClass.getURI().equalsIgnoreCase(
								OWL.Nothing.getURI())) {
							// check that current node exist. if not, create
							// Node
							// and register it
							if (!alreadyFoundConcepts.containsKey(subClass
									.getURI())) {
								Node currenSubtNode = new Node(subClass
										.getURI());
								alreadyFoundConcepts.put(subClass.getURI(),
										currenSubtNode);
								//DebugPrinter.print("put "+currentClass.getURI()+" in alreadyFoundConcept" , DebugPrinter.INFO);								
								classesToCheck.add(subClass);
							}
						}
					} /*
						 * else { DebugPrinter.print("Something wrong with a
						 * children of " + currentClass.getURI()+"
						 * ("+subClass.getLocalName()+")", DebugPrinter.WARN); }
						 */
				}

				/*
				 * } else { DebugPrinter.print("Something wrong with " +
				 * currentClass, DebugPrinter.WARN); }
				 */
			}
			/*
			 * } catch (NullPointerException e) { DebugPrinter.print(e,
			 * DebugPrinter.WARN); }
			 */
		}
	}

}
